import datetime
import semantic_version
import cnr.packager as packager
from cnr.semver import last_version, select_version
from cnr.exception import (InvalidVersion,
                           PackageAlreadyExists,
                           ChannelNotFound,
                           PackageVersionNotFound,
                           PackageNotFound)


class PackageBase(object):
    def __init__(self, package_name, version=None, blob=None, data=None):
        self.package = package_name
        self.namespace, self.name = package_name.split("/")
        self.version = version
        self._data = data
        self.created_at = None
        self.packager = None
        self.blob = blob

    def channels(self, channel_class):
        """ Returns all available channels for a package """
        channel_names = channel_class.all(self.package)
        result = []
        for channel in channel_names:
            c = channel_class(channel, self.package)
            releases = c.releases()
            if self.version in releases:
                result.append(channel)
        return result

    @property
    def blob(self):
        return self.packager.b64blob

    @property
    def digest(self):
        return self.packager.digest

    @blob.setter
    def blob(self, value):
        if value is not None:
            self.packager = packager.Package(value)

    @property
    def data(self):
        if self._data is None:
            self._data = {'created_at': datetime.datetime.utcnow().isoformat()}
        d = {"package": self.package,
             "release": self.version,
             "blob": self.blob,
             "digest": self.packager.digest}

        self._data.update(d)
        return self._data

    @data.setter
    def data(self, data):
        self._data = data
        self.blob = data['blob']
        self.created_at = data['created_at']
        self.version = data['release']

    @classmethod
    def check_version(self, version):
        try:
            semantic_version.Version(version)
        except ValueError as e:
            raise InvalidVersion(e.message, {"version": version})
        return None

    @classmethod
    def get(self, package, version='default'):
        """
        package: string following "namespace/package_name" format
        version: version query. If None return latest version

        returns: (package blob(targz) encoded in base64, version)
        """
        p = self(package, version)
        p.pull(version)
        return p

    @classmethod
    def get_version(self, package, version_query, stable=False):
        versions = self.all_versions(package)
        if not versions:
            self._raise_not_found(package)
        if version_query is None or version_query == 'default':
            return last_version(versions, stable)
        else:
            try:
                return select_version(versions, str(version_query), stable)
            except ValueError as e:
                raise InvalidVersion(e.message, {"version": version_query})

    def pull(self, version_query=None):
        if version_query is None:
            version_query = self.version
        package = self.package
        version = self.get_version(package, version_query)
        if version is None:
            raise PackageVersionNotFound("No version match '%s' for package '%s'" % (version_query, package),
                                         {"package": package, "version_query": version_query})

        self.data = self._fetch(package, version)
        return self

    @classmethod
    def isdeleted_release(self, package, version):
        raise NotImplementedError

    @classmethod
    def search_index(self):
        pass

    @classmethod
    def add_index(self, name):
        pass

    @classmethod
    def remove_index(self, name):
        pass

    @classmethod
    def write_index(self, index):
        pass

    @classmethod
    def reindex(self):
        r = set()
        for package in self.all():
            r.add(package['name'])
        self.write_index(r)

    def save(self, force=False):
        self.check_version(self.version)
        if self.isdeleted_release(self.package, self.version) and not force:
            raise PackageAlreadyExists("Package release %s existed" % self.package,
                                       {"package": self.package, "version": self.version})
        self._save(force)
        self.add_index(self.package)

    def versions(self):
        return self.all_versions(self.package)

    @classmethod
    def _raise_not_found(self, package, version=None):
        raise PackageNotFound("package %s doesn't exist" % package,
                              {'package': package, 'version': version})

    @classmethod
    def all(self, namespace=None):
        raise NotImplementedError

    @classmethod
    def _fetch(self, package, version):
        raise NotImplementedError

    def _save(self, force=False):
        raise NotImplementedError

    @classmethod
    def _delete(self, package, version):
        raise NotImplementedError

    @classmethod
    def _delete_from_channels(self, package, version, channel_class):
        p = self(package, version)
        for channel in p.channels(channel_class):
            c = channel_class(channel, package)
            try:
                c.remove_release(version)
            except ChannelNotFound:
                pass

    @classmethod
    def delete(self, package, version, channel_class):
        self._delete_from_channels(package, version, channel_class)
        self._delete(package, version)
        if not self.all_versions(package):
            self.remove_index(package)

    @classmethod
    def all_versions(self, package):
        raise NotImplementedError

    @classmethod
    def search(self, query):
        raise NotImplementedError



from flask import (jsonify,
                   request,
                   Blueprint,
                   current_app,
                   url_for)

import cnr


info_app = Blueprint('info', __name__,)


@info_app.before_request
def pre_request_logging():
    jsonbody = request.get_json(force=True, silent=True)
    values = request.values.to_dict()
    if jsonbody:
        values.update(jsonbody)

    current_app.logger.info("request", extra={
        "remote_addr": request.remote_addr,
        "http_method": request.method,
        "original_url": request.url,
        "path": request.path,
        "data":  values,
        "headers": dict(request.headers.to_list())})


@info_app.route("/")
def index_discovery():
    host = request.url_root
    domain = request.headers['Host']
    return """<html lang="en">
    <head>
    <meta charset="utf-8">
    <meta name="viewport" content="width=device-width, initial-scale=1.0">
    <meta name="cnr-package" content="{domain}/{{name}} {host}/cnr/api/v1/packages/{{name}}/pull">
    </head>
    <body>
    </body>
    </html>""".format(domain=domain, host=host)


@info_app.route("/version")
def version():
    return jsonify({"cnr-api": cnr.__version__})


@info_app.route("/routes")
def routes():
    import urllib
    output = []
    for rule in current_app.url_map.iter_rules():
        options = {}
        for arg in rule.arguments:
            options[arg] = "[{0}]".format(arg)
        methods = ','.join(rule.methods)
        url = url_for(rule.endpoint, **options)
        line = urllib.unquote("{:50s} {:20s} {}".format(rule.endpoint, methods, url))
        output.append(line)
    lines = []
    for line in sorted(output):
        lines.append(line)
    return jsonify({"routes": lines})



import json
import pytest
import os
from cnr.api.app import create_app


@pytest.fixture
def discovery_html():
    return """<html lang="en">
    <head>
    <meta charset="utf-8">
    <meta name="viewport" content="width=device-width, initial-scale=1.0">
    <meta name="cnr-package" content="cnr.sh/{name} https://api.kubespray.io/api/v1/packages/{name}/pull">
    </head>
    <body>
    <a href=https://github.com/coreos/cnr>coreos/cnr</a>
    </body>
    </html>"""


@pytest.fixture
def app():
    app = create_app()
    return app


@pytest.fixture()
def fake_home(monkeypatch, tmpdir):
    home = tmpdir.mkdir('home')
    monkeypatch.setenv("HOME", home)
    return home


def get_response(name, kind):
    f = open("tests/data/responses/%s-%s.json" % (name, kind))
    r = f.read()
    f.close()
    return r


@pytest.fixture(scope="module")
def kubeui_package():
    import base64
    import cnr.packager
    with open("./tests/data/kube-ui.tar.gz", "rb") as f:
        package = cnr.packager.Package(base64.b64encode(f.read()))
    return package


@pytest.fixture(scope='module')
def package_data():
    return 'H4sICF8BsVcC/3RpdF9yb2NrZXRjaGF0XzEuMTAuMGt1Yi50YXIA7Zddb9MwFIZ7nV9hlYve0DRpmgZFgJjGhCat2zQGNwhNXup1Vu04st1IVZX/znGTtMnYWAWj08DPRT6Ofezjr/ckF0cHHydHLp92/h4eMB6PzN2PQq95L4lGHX84HkfjcBz4YPeDKPA6yOvsgYXSWCLUwYzMJE4fnIfHyquhbO4vBEeKZE50cou1826L47xCxylMDWOOM884ouULalR3OpYXD8cpvSFKu0vM2fOcf98LAjj/URgE/ggK4PyHPpjs+d8D/X7fyXAyxzMSOwilmJMYdTXVg+1J70IBXuhbIWN0kGpBU4JOytlAb4faxaXtw4xjytxE8PfgkBOpqEhj5Lu+53pgmRKVSJrptbWhIwgxmpBUQceT40vHybGk+JoRZQKSJINSrKAdeKPcxNlwrqJ0zXPMsIadXI1CwajInX5MfFcLyWKkuM7iwQBsCYMdQGBk1YPLRIKZqSzSmZhel/WrF3CpntzVatNLUbgqT8BSNVEU8TCCLT5oiqUkSixkUo6qj24oa0XXNy0suem4XoVW6AjpZQZGqObc34DcyV+C+xTmVCzLOMqqoPCuuh3AQgZ+PUIr7/8DmvBsfW4GP22lvel/4Jnvv8gLQ89cQP+jyB9a/d+X/uOMfq3VOvedOU2nMboohdeo9aFItRSMEelwovEUa7zNFW0hx9eEgVavevM3qo+zrBej3rZG7zXqVXnBFOR+r2irdUtUHZWRpJ0DVqv6uTCeijCSaJOWHurQ1Kq3eLyWweYIDL8VMnRdxWZIYH4wJECpakv/fgk27JjCSlpJo4ZRTnXLAiFkC8iQnsdbVk64kMsYBZ43oZsSkuZN5zrWydnpp7OrLxcnrSZyzBbrhWmkw6K4z/3g+OQX3lXmbbhmQsIo0LdVL5NCi0QwM8GXh+dmyjdTeg61eusReGA2Xd1d4e82UT21/lffIk+q/6OH//9Hw9D8/0ejcBSGQ7AP4Z9gbPX/OfS/m/vdKgN0PxOZw5d5909VfyeR31HL1x+xp2JKjDA4j6gIrOyM6EclBKzZto4VFIvFYrFYLBaLxWKxWCwWi8Xyb/ADcPSUWQAoAAA='


@pytest.fixture(scope="module")
def kubeui_blob():
    import cnr.packager
    with open("./tests/data/kube-ui.tar.gz", "rb") as f:
        package = f.read()
    return package


@pytest.fixture()
def package_dir(monkeypatch):
    monkeypatch.chdir("tests/data/kube-ui")


@pytest.fixture()
def bad_package_dir(monkeypatch):
    monkeypatch.chdir("tests/data/bad_manifest")

@pytest.fixture()
def empty_package_dir(monkeypatch):
    monkeypatch.chdir("tests/data")


@pytest.fixture()
def pack_tar(package_dir, tmpdir):
    from cnr.packager import pack_kub
    kub = os.path.join(str(tmpdir.mkdir("tars")), "kube-ui.tar.gz")
    pack_kub(kub)
    return kub


@pytest.fixture(scope="module")
def deploy(deploy_json):
    return json.loads(deploy_json)


@pytest.fixture(scope="module")
def ns_resource(deploy):
    kubeui = deploy["deploy"][0]
    return kubeui['resources'][0]


@pytest.fixture(scope="module")
def rc_resource(deploy):
    kubeui = deploy["deploy"][0]
    return kubeui['resources'][1]


@pytest.fixture(scope="module")
def svc_resource(deploy):
    kubeui = deploy["deploy"][0]
    return kubeui['resources'][2]



import json
from base64 import b64decode
from flask import jsonify, request, Blueprint, current_app
from cnr.api.app import getvalues
import cnr.api.impl.registry
from cnr.exception import (CnrException,
                           InvalidUsage,
                           InvalidVersion,
                           PackageAlreadyExists,
                           ChannelAlreadyExists,
                           PackageNotFound,
                           ChannelNotFound,
                           PackageVersionNotFound)

# @TODO chose data model from configuration
from cnr.models.etcd.package import Package
from cnr.models.etcd.channel import Channel


registry_app = Blueprint('registry', __name__,)


@registry_app.errorhandler(PackageAlreadyExists)
@registry_app.errorhandler(ChannelAlreadyExists)
@registry_app.errorhandler(InvalidVersion)
@registry_app.errorhandler(PackageNotFound)
@registry_app.errorhandler(PackageVersionNotFound)
@registry_app.errorhandler(CnrException)
@registry_app.errorhandler(InvalidUsage)
@registry_app.errorhandler(ChannelNotFound)
def render_error(error):
    response = jsonify({"error": error.to_dict()})
    response.status_code = error.status_code
    return response


@registry_app.route("/test_error")
def test_error():
    raise InvalidUsage("error message", {"path": request.path})


@registry_app.route("/api/v1/packages/<path:package>/pull", methods=['GET'], strict_slashes=False)
def pull(package):
    values = getvalues()
    version = values.get("version", None)
    data = cnr.api.impl.registry.pull(package, version, Package)
    if 'format' in values and values['format'] == 'json':
        resp = jsonify({"package": data['package'], "blob": data['blob']})
    else:
        resp = current_app.make_response(b64decode(data['blob']))
        resp.headers['Content-Disposition'] = data['filename']
        resp.mimetype = 'application/x-gzip'
    return resp


@registry_app.route("/api/v1/packages", methods=['POST'], strict_slashes=False)
def push():
    values = getvalues()
    blob = values['blob']
    package = values['package']
    version = values['version']
    force = False
    if 'force' in values:
        force = 'true' == values['force']

    r = cnr.api.impl.registry.push(package, version, blob, force, Package)
    return jsonify(r)


@registry_app.route("/api/v1/packages", methods=['GET'], strict_slashes=False)
def list_packages():
    values = getvalues()
    namespace = values.get('namespace', None)
    r = cnr.api.impl.registry.list_packages(namespace, Package)
    resp = current_app.make_response(json.dumps(r))
    resp.mimetype = 'application/json'
    return resp


@registry_app.route("/api/v1/packages/search", methods=['GET'], strict_slashes=False)
def search_packages():
    values = getvalues()
    query = values.get("q")
    r = cnr.api.impl.registry.search(query, Package)
    return jsonify(r)


@registry_app.route("/api/v1/packages/search_reindex", methods=['POST'], strict_slashes=False)
def search_reindex():
    r = cnr.api.impl.registry.search_reindex(Package)
    return jsonify(r)


@registry_app.route("/api/v1/packages/<path:package>", methods=['GET'], strict_slashes=False)
def show_package(package):
    values = getvalues()
    version = values.get("version", 'default')
    r = cnr.api.impl.registry.show_package(package, version,
                                           channel_class=Channel,
                                           package_class=Package)
    return jsonify(r)


# CHANNELS
@registry_app.route("/api/v1/packages/<path:package>/channels", methods=['GET'], strict_slashes=False)
def list_channels(package):
    r = cnr.api.impl.registry.list_channels(package, Channel)
    resp = current_app.make_response(json.dumps(r))
    resp.mimetype = 'application/json'
    return resp


@registry_app.route("/api/v1/packages/<path:package>/channels/<string:name>", methods=['GET'], strict_slashes=False)
def show_channel(package, name):
    r = cnr.api.impl.registry.show_channel(package, name, Channel)
    return jsonify(r)


@registry_app.route("/api/v1/packages/<path:package>/channels/<string:name>/<string:release>",
                    methods=['POST'], strict_slashes=False)
def add_channel_release(package, name, release):
    r = cnr.api.impl.registry.add_channel_release(package, name, release,
                                                  channel_class=Channel,
                                                  package_class=Package)
    return jsonify(r)


@registry_app.route("/api/v1/packages/<path:package>/channels/<string:name>/<string:release>",
                    methods=['DELETE'], strict_slashes=False)
def delete_channel_release(package, name, release):
    r = cnr.api.impl.registry.delete_channel_release(package, name, release,
                                                     channel_class=Channel,
                                                     package_class=Package)
    return jsonify(r)


@registry_app.route("/api/v1/packages/<path:package>/channels/<string:name>",
                    methods=['POST'], strict_slashes=False)
def create_channel(package, name):
    r = cnr.api.impl.registry.create_channel(package, name, Channel)
    return jsonify(r)


@registry_app.route("/api/v1/packages/<path:package>/channels/<string:name>",
                    methods=['DELETE'], strict_slashes=False)
def delete_channel(package, name):
    r = cnr.api.impl.registry.delete_channel(package, name,
                                             channel_class=Channel)
    return jsonify(r)


@registry_app.route("/api/v1/packages/<string:namespace>/<string:name>", methods=['DELETE'], strict_slashes=False)
def delete_package(namespace, name):
    package = "%s/%s" % (namespace, name)
    values = getvalues()
    version = values.get("version", "default")
    r = cnr.api.impl.registry.delete_package(package, version,
                                             package_class=Package,
                                             channel_class=Channel)
    return jsonify(r)



import base64
import gzip
import hashlib
import logging
import tarfile
import io
import os


LOGGER = logging.getLogger(__name__)


def authorized_files():
    files = []
    for root, _, filenames in os.walk('.'):
        for filename in filenames:
            files.append(os.path.join(root, filename))
    return files


def pack_app(app):
    tar = tarfile.open(app, "w:gz")
    for filename in authorized_files():
        tar.add(filename)
    tar.close()


def unpack_app(app, dest="."):
    tar = tarfile.open(app, "r:gz")
    tar.extractall(dest)
    tar.close()


class Package(object):
    def __init__(self, blob=None, b64_encoded=True):
        self.files = {}
        self.tar = None
        self.blob = None
        self.io_file = None
        self._digest = None
        self.b64blob = None
        if blob is not None:
            self.load(blob, b64_encoded)

    def _load_blob(self, blob, b64_encoded):
        if b64_encoded:
            self.b64blob = blob
            self.blob = base64.b64decode(blob)
        else:
            self.b64blob = base64.b64encode(blob)
            self.blob = blob

    def load(self, blob, b64_encoded=True):
        self._digest = None
        self._load_blob(blob, b64_encoded)
        self.io_file = io.BytesIO(self.blob)
        self.tar = tarfile.open(fileobj=self.io_file, mode='r:gz')
        for member in self.tar.getmembers():
            tfile = self.tar.extractfile(member)
            if tfile is not None:
                self.files[tfile.name] = tfile.read()

    def extract(self, dest):
        self.tar.extractall(dest)

    def pack(self, dest):
        with open(dest, "wb") as destfile:
            destfile.write(self.blob)

    def tree(self, directory=None):
        files = self.files.keys()
        files.sort()
        if directory is not None:
            filtered = [x for x in files if x.startswith(directory)]
        else:
            filtered = files
        return filtered

    def file(self, filename):
        return self.files[filename]

    @property
    def digest(self):
        if self._digest is None:
            self.io_file.seek(0)
            gunzip = gzip.GzipFile(fileobj=self.io_file, mode='r').read()
            self._digest = hashlib.sha256(gunzip).hexdigest()
        return self._digest



import datetime
from cnr.exception import ChannelNotFound, ChannelAlreadyExists
import cnr.semver as semver


class ChannelBase(object):
    def __init__(self, name, package):
        self.package = package
        self.name = name
        self._iscreated = None

    def exists(self):
        return self._exists()

    def _exists(self):
        """ Check if the channel is saved already """
        raise NotImplementedError

    def save(self):
        raise NotImplementedError

    def delete(self):
        raise NotImplementedError

    @classmethod
    def all_channels(cls, package):
        """ Returns all available channels for a package """
        channel_names = cls.all(package)
        result = {}
        for channel_name in channel_names:
            channel = cls(channel_name, package)
            releases = channel.releases()
            result[str(channel)] = {"releases": releases,
                                    "name": channel_name,
                                    "current": channel.current_release(releases)}
        return result

    @classmethod
    def all(cls, package):
        raise NotImplementedError

    def releases(self):
        """ Returns the list of versions """
        raise NotImplementedError

    def _add_release(self, version):
        # etcdctl put /{self.package/channels/{self.name}/version
        raise NotImplementedError

    def _remove_release(self, version):
        # etcdctl put /{self.package/channels/{self.name}/version
        raise NotImplementedError

    def current_release(self, releases=None):
        if releases is None:
            releases = self.releases()
        if not releases:
            return None
        ordered_versions = [str(x) for x in sorted(semver.versions(releases, False),
                                                   reverse=True)]
        return ordered_versions[0]

    def add_release(self, version, package_class):
        if self._check_release(version, package_class) is False:
            raise ValueError("Release %s doesn't exist for package %s" % (version, self.package))
        if not self.exists():
            self.save()
        return self._add_release(version)

    def remove_release(self, version):
        if not self.exists():
            self.save()
        return self._remove_release(version)

    def _check_release(self, release_name, package_class):
        version = package_class.get_version(self.package, release_name)
        if version is None or str(version) != release_name:
            return False
        else:
            return True

    def _new_chan_release(self, version):
        data = {'release': version,
                'created_at': datetime.datetime.utcnow().isoformat()}
        return data

    @classmethod
    def _raise_not_found(cls, package, channel=None, version=None):
        if channel is None:
            raise ChannelNotFound("No channel found for package '%s'" % (package),
                                  {'package': package})
        else:
            raise ChannelNotFound("Channel '%s' doesn't exist for package '%s'" % (channel, package),
                                  {'channel': channel, 'package': package, 'version': version})

    @classmethod
    def _raise_already_exists(cls, channel, package, version=None):
        if version is None:
            raise ChannelAlreadyExists("Channel '%s' exists already for package '%s'" % (channel, package),
                                       {'package': package, 'channel': channel})
        else:
            raise ChannelAlreadyExists("Realease '%s' exists already in channel '%s'" % (version, channel),
                                       {'package': package, 'channel': channel, 'version': version})

    def to_dict(self):
        releases = self.releases()
        return ({"releases": releases,
                 "name": self.name,
                 "current": self.current_release(releases)})



import os
from flask import Flask, request
from flask_cors import CORS


def getvalues():
    jsonbody = request.get_json(force=True, silent=True)
    values = request.values.to_dict()
    if jsonbody:
        values.update(jsonbody)
    return values


def create_app():
    app = Flask(__name__)
    CORS(app)
    setting = os.getenv('APP_ENV', "development")

    if setting != 'production':
        app.config.from_object('cnr.api.config.DevelopmentConfig')
    else:
        app.config.from_object('cnr.api.config.ProductionConfig')
    from cnr.api.info import info_app
    from cnr.api.registry import registry_app
    app.register_blueprint(info_app, url_prefix='/cnr')
    app.register_blueprint(registry_app, url_prefix='/cnr')
    app.logger.info("Start service")
    return app


if __name__ == "__main__":
    application = create_app()
    application.run(host='0.0.0.0')



import os


class Config(object):
    """ Default configuration """
    DEBUG = False
    CNR_URI = os.getenv('CNR_URI', "http://localhost:5000")


class ProductionConfig(Config):
    """ Production configuration """
    CNR_URI = "http://localhost:5000"
    CNR_BACKEND = 'false'


class DevelopmentConfig(Config):
    """ Development configuration """
    DEBUG = True
#    CNR_URI = 'https://api.cnr.sh'
    CNR_URI = os.getenv('CNR_URI', "http://localhost:5000")



import datetime
import json
import etcd
import re
import cnr.semver as semver
from cnr.models.package_base import PackageBase
from cnr.exception import PackageAlreadyExists
from cnr.models.etcd import ETCD_PREFIX, etcd_client


class Package(PackageBase):
    index_path = "/cnr/search/index"

    def __init__(self, package_name, version=None, blob=None):

        super(Package, self).__init__(package_name,
                                      version=version,
                                      blob=blob)

    @classmethod
    def _fetch(self, package, version):
        path = self._etcdkey(package, str(version))
        try:
            package_data = json.loads(etcd_client.read(path).value)
        except etcd.EtcdKeyNotFound:
            self._raise_not_found(package, version)
        return package_data

    @classmethod
    def all_versions(self, package):
        path = ETCD_PREFIX + package + "/releases"
        try:
            r = etcd_client.read(path, recursive=True)
        except etcd.EtcdKeyNotFound:
            self._raise_not_found(package, None)

        versions = []
        for child in r.children:
            m = re.match("^/%s(.+)/(.+)/releases/(.+)$" % ETCD_PREFIX, child.key)
            if m is None:
                continue
            version = m.group(3)
            versions.append(version)
        return versions

    @classmethod
    def deleted_releases(self, package):
        path = ETCD_PREFIX + "%s/deleted" % package
        r = []
        try:
            deleted_releases = etcd_client.get(path)
            for child in deleted_releases.children:
                m = re.match("^/%s/(.+)$" % path, child.key)
                if m is None:
                    continue
                r.append(m.group(1))
        except etcd.EtcdKeyNotFound:
            pass
        return r

    @classmethod
    def all(self, namespace=None):
        path = ETCD_PREFIX
        r = {}
        if namespace is not None:
            path += "/%s" % namespace
        try:
            packages = etcd_client.read(path, recursive=True)
        except etcd.EtcdKeyNotFound:
            etcd_client.write(path, None, dir=True)
            packages = etcd_client.read(path, recursive=True)

        for child in packages.children:
            m = re.match("^/%s(.+)/(.+)/releases/(.+)$" % ETCD_PREFIX, child.key)
            if m is None:
                continue
            namespace, name, version = (m.group(1), m.group(2), m.group(3))
            package = "%s/%s" % (namespace, name)
            if package not in r:
                r[package] = {"name": package, 'available_releases': [], 'version': None}
            r[package]['available_releases'].append(version)
        for _, v in r.iteritems():
            v['available_releases'] = [str(x) for x in sorted(semver.versions(v['available_releases'], False),
                                                              reverse=True)]
            v['version'] = v['available_releases'][0]
        return r.values()

    @classmethod
    def isdeleted_release(self, package, version):
        if version in self.deleted_releases(package):
            return True
        else:
            return False

    def _save(self, force=False):
        v = self._push_etcd(self.package, self.version, json.dumps(self.data), force=force)
        try:
            digest_path = self._etcdkey(self.package, self.data['digest'], 'digests')
            etcd_client.write(digest_path, str(v.key))
        except etcd.EtcdAlreadyExist as e:
            raise PackageAlreadyExists(e.message, {"package": self.package, "version": self.version})

    @classmethod
    def search_index(self):
        try:
            r = json.loads(etcd_client.get(self.index_path).value)
        except etcd.EtcdKeyNotFound:
            r = []
            self.write_index(r)
        return r

    @classmethod
    def add_index(self, name):
        r = set(self.search_index())
        r.add(name)
        self.write_index(r)

    @classmethod
    def write_index(self, r):
        etcd_client.write(self.index_path, json.dumps(list(r)))

    @classmethod
    def remove_index(self, name):
        r = set(self.search_index())
        r.remove(name)
        self.write_index(r)

    @classmethod
    def search(self, query):
        r = '\n'.join(self.search_index())
        return re.findall(r"(.*%s.*)" % query, r)

    @classmethod
    def _delete(self, package, version):
        path = self._etcdkey(package, str(version))
        try:
            self._fetch(package, version)
            etcd_client.write(path.replace("releases", "deleted"),
                              json.dumps({"deleted_at": datetime.datetime.utcnow().isoformat()}))
            etcd_client.delete(path)
        except etcd.EtcdKeyNotFound:
            self._raise_not_found(package, version)

    @classmethod
    def _etcdkey(self, package, key, directory='releases'):
        return ETCD_PREFIX + "%s/%s/%s" % (package, directory, key)

    def _push_etcd(self, package, version, data, force=False):
        path = self._etcdkey(package, version)
        try:
            r = etcd_client.write(path, data, prevExist=force)
        except etcd.EtcdAlreadyExist as e:
            raise PackageAlreadyExists(e.message, {"package": path})
        except etcd.EtcdKeyNotFound as e:
            r = etcd_client.write(path, data, prevExist=False)
        return r



import logging
import cnr.semver as semver


logger = logging.getLogger(__name__)


def _get_package(package, version_query, package_class):
    """
      Fetch the package data from the datastore
      and instantiate a :obj:`cnr.models.package_base.PackageModelBase`

    Args:
      package (:obj:`str`): package name in the format "namespace/name" or "domain.com/name"
      version_query (:obj:`str`): a version query, eg: ">=1.5,<2.0"
      package_class (:obj:`cnr.models.package_base:PackageBase`): the implemented Package class to use

    Returns:
      :obj:`cnr.kub_jsonnet.KubJsonnet`: :obj:`cnr.models.package_base.PackageModelBase`

    See Also:
       * :obj:`cnr.api.models.package_base.PackageModelBase`
       * :obj:`cnr.api.models.etcd.package.Package`

    Raises:
      :obj:`cnr.exception.PackageNotFound`: package not found
      :obj:`cnr.exception.InvalidVersion`: version-query malformated
      :obj:`cnr.exception.PackageVersionNotFound`: version-query didn't match any release

    """

    if version_query is None:
        version_query = 'default'
    p = package_class.get(package, version_query)
    return p


def pull(package, version, package_class):
    """
    Retrives the package blob from the datastore

    Args:
      package (:obj:`str`): package name in the format "namespace/name" or "domain.com/name"
      version_query (:obj:`str`): a version query, eg: ">=1.5,<2.0"
      package_class (:obj:`cnr.models.package_base:PackageBase`): the implemented Package class to use

    Returns:
      :obj:`dict`: package data
        * package: package name
        * version: version that matched the version query
        * filename: suggested filename to create the tarball
        * blob: a `tar.gz` encoded in base64.

    Example:
      >>> cnr.api.impl.registry.pull("coreos/etcd", version=">=3")
        {
         'blob': 'H4sICHDFvlcC/3RpdF9yb2NrZXRjaGF0XzEuMTAuMGt1Yi50YXIA7ZdRb5swEM',
         'filename': u'coreos_etcd_3.0.1.tar.gz',
         'package': 'coreos_etcd',
         'version': u'3.0.1'
        }

    Raises:
      :obj:`cnr.exception.PackageNotFound`: package not found
      :obj:`cnr.exception.InvalidVersion`: version-query malformated
      :obj:`cnr.exception.PackageVersionNotFound`: version-query didn't match any release

    See Also:
       * :obj:`cnr.api.registry.pull`

    """

    packagemodel = _get_package(package, version, package_class=package_class)
    resp = {"package": package,
            "blob": packagemodel.blob,
            "version": packagemodel.version,
            "filename": "%s_%s.tar.gz" % (packagemodel.package.replace("/", "_"), packagemodel.version)}
    return resp


def push(package, version, blob, force, package_class):
    """
    Push a new package release in the the datastore

    Args:
      package (:obj:`str`): package name in the format "namespace/name" or "domain.com/name"
      version (:obj:`str`): the 'exact' package version (this is not a version_query)
      blob (:obj:`str`): the package directory in `tar.gz` and encoded in base64
      force (:obj:`boolean`): if the package exists already, overwrite it
      package_class (:obj:`cnr.models.package_base:PackageBase`): the implemented Package class to use

    Returns:
      :obj:`dict`: push status

    Example:
      >>> cnr.api.impl.registry.push("coreos/etcd", "3.0.1",
            "H4sICHDFvlcC/3RpdF9yb2NrZXRjaGF0XzEuMTAuMGt1Yi50YXIA7ZdRb5swEM")
        {
         'status': u'ok'
        }

    Raises:
      PackageAlreadyExists: if package already exists and `force` is False

    See Also:
       * :obj:`cnr.api.registry.push`

    """
    p = package_class(package, version, blob)
    p.save(force=force)
    return {"status": "ok"}


def search_reindex(package_class):
    package_class.reindex()
    return {"status": "ok"}


def search(query, package_class):
    """
    Search packages

    Args:
      package (:obj:`str`): package name in the format "namespace/name" or "domain.com/name"
      package_class (:obj:`cnr.models.package_base:PackageBase`): the implemented Package class to use

    Returns:
      :obj:`list`: list of package names

    Example:
      >>> cnr.api.impl.registry.search("etcd")
        [
         'coreos/etcd',
         'ant31/etcd',
         'my-etcd/stable'
        ]

    See Also:
       * :obj:`cnr.api.registry.search`

    """
    p = package_class.search(query)
    return p


def list_packages(namespace, package_class):
    """
    List all packages, filters can be applied
    Must have at least a release to be visible

    Todos:
       - sort_by: name, created_at, downloads, number of stars
       - filter_by: users

    Args:
      namespace (:obj:`str`): returns packages from the `namespace` only
      package_class (:obj:`cnr.models.package_base:PackageBase`): the implemented Package class to use

    Returns:
      :obj:`list of dict`: list packages
        * name: package name
        * available_releases (list of str):  All releases
        * created_at (datetime, optional): package creation date
        * downloads (int, optional): number of downloads
        * version: release name

    Example:
      >>> cnr.api.impl.registry.list_packages()
      [
       {
        'available_releases': ['0.1.0'],
        'name': u'quentinm/rados-gateway',
        'version': '0.1.0',
        'created_at": "2016-04-22T11:58:34.103Z",
        'downloads': 41
       },
       {
        'available_releases': ['0.1.0'],
        'name': u'quentinm/nova',
        'version': '0.1.0'
       },
      ]

    See Also:
       * :obj:`cnr.api.registry.list_packages`
    """
    resp = package_class.all(namespace)
    return resp


def show_package(package,
                 version,
                 channel_class,
                 package_class):
    """
    Returns package details

    Args:
      package (:obj:`str`): package name in the format "namespace/name" or "domain.com/name"
      version (:obj:`str`): the 'exact' package version (this is not a version_query)
      channel_class (:obj:`cnr.models.channel_base:ChannelBase`): the implemented Channel class to use
      package_class (:obj:`cnr.models.package_base:PackageBase`): the implemented Package class to use

    Returns:
      :obj:`dict`: package data
        * version (str)
        * name (str)
        * created_at (str)
        * digest (str)
        * channels (list)
        * available_releases (list)
        * dependencies (list)
        * variables (dict)
        * manifest (str)

    Example:
      >>> cnr.api.impl.registry.show_package("ns/mypackage")
      {
      "version": "3.2.0-rc",
      "name": "ns/mypackage",
      "created_at": "2016-08-25T10:16:16.366758",
      "digest": "93de60f59238f9aebce5b9f8bc708e02a0d740364fcd4b185c6da7fc1cdfe1ba",
      "channels": ['stable', 'beta'],
      "available_releases": [
        "3.2.0-rc"
        "3.1.0",
        "3.0.1"
      ],
      "dependencies": [
        "ns/dep1",
        "ns/dep2",
        "ns/dep3"
       ],
       "variables": {
         "replicas": 1,
         "image": "ns/mypackage:latest",
         "namespace": "default",
         "cluster": "cluster.local",
         "mail_url": "smtp://"
       },
      "manifest": "---...."
       }

    Raises:
      :obj:`cnr.exception.PackageNotFound`: package not found
      :obj:`cnr.exception.InvalidVersion`: version-query malformated
      :obj:`cnr.exception.PackageVersionNotFound`: version-query didn't match any release

    See Also:
       * :obj:`cnr.api.registry.show_package`
    """
    stable = False
    packagemodel = _get_package(package, version, package_class)
    manifest = packagemodel.manifest()
    response = {"manifest": packagemodel.packager.manifest,
                "version": packagemodel.version,
                "name":  package,
                "created_at": packagemodel.created_at,
                "digest": packagemodel.digest,
                "variables": manifest.variables,
                "dependencies": manifest.dependencies,
                "channels": packagemodel.channels(channel_class),
                "available_releases": [str(x) for x in sorted(semver.versions(packagemodel.versions(), stable),
                                                              reverse=True)]}
    return response


# CHANNELS
def list_channels(package, channel_class):
    """
    List all channels for a given package

    Args:
      package (:obj:`str`): package name in the format "namespace/name" or "domain.com/name"
      channel_class (:obj:`cnr.models.channel_base:ChannelBase`): the implemented Channel class to use

    Returns:
      :obj:`list of dict`: list channels:
        * channel (str): channel name
        * current (str): latest/default release associated to the channel
        * releases (list): list channel's releases

    Example:
      >>> cnr.api.impl.registry.list_channels("myns/package")
      [{'channel': u'stable', 'current': '1.10.2', 'releases': [u'1.10.2']},
       {'channel': u'dev', 'current': 2.0.0-beta, 'releases': [1.10.2, 2.0.0-beta]}]

    See Also:
       * :obj:`cnr.api.registry.list_channels`
    """
    channels = channel_class.all_channels(package).values()
    return channels


def show_channel(package, name, channel_class):
    """
    Show channel info
    Args:
      package (:obj:`str`): package name in the format "namespace/name" or "domain.com/name"
      name (:obj:`str`): channel name to inspect
      channel_class (:obj:`cnr.models.channel_base:ChannelBase`): the implemented Channel class to use

    Returns:
      :obj:`dict`: channel info
        * channel (str): channel name
        * current (str): latest/default release associated to the channel
        * releases (list): list channel's releases

    Example:
      >>> cnr.api.impl.registry.list_channels("tit/rocketchat", 'dev')
      {'channel': u'dev', 'current': '2.0.0-beta', 'releases': [u'1.10.2']}

    Raises:
      :obj:`cnr.api.exception.ChannelNotFound`: channel not found

    See Also:
       * :obj:`cnr.api.registry.show_channel`
    """
    c = channel_class(name, package)
    return c.to_dict()


def add_channel_release(package, name, release, channel_class, package_class):
    """
    Add a package-release to a channel
    Args:
      package (:obj:`str`): package name in the format "namespace/name" or "domain.com/name"
      name (:obj:`str`): channel name to inspect
      release (:obj:`str`): package version to add
      channel_class (:obj:`cnr.models.channel_base:ChannelBase`): the implemented Channel class to use
      package_class (:obj:`cnr.models.package_base:PackageBase`): the implemented Package class to use

    Returns:
      :obj:`dict`: channel info
        * channel (str): channel name
        * current (str): latest/default release associated to the channel
        * releases (list): list channel's releases

    Example:
      >>> cnr.api.impl.registry.list_channels("tit/rocketchat", 'dev')
      {'channel': u'dev', 'current': '2.0.0-beta', 'releases': [u'1.10.2']}

    Raises:
      :obj:`cnr.api.exception.ChannelNotFound`: channel not found

    See Also:
       * :obj:`cnr.api.registry.add_channel_release`
    """
    channel = channel_class(name, package)
    channel.add_release(release, package_class)
    return channel.to_dict()


def delete_channel_release(package, name, release, channel_class, package_class):
    """
    Remove a release from a channel

    Args:
      package (:obj:`str`): package name in the format "namespace/name" or "domain.com/name"
      name (:obj:`str`): channel name to inspect
      release (:obj:`str`): package version to add
      channel_class (:obj:`cnr.models.channel_base:ChannelBase`): the implemented Channel class to use
      package_class (:obj:`cnr.models.package_base:PackageBase`): the implemented Package class to use

    Returns:
      :obj:`dict`: channel info
        * channel (str): channel name
        * current (str): latest/default release associated to the channel
        * releases (list): list channel's releases

    Example:
      >>> cnr.api.impl.registry.list_channels("tit/rocketchat", 'dev')
      {'channel': u'dev', 'current': '2.0.0-beta', 'releases': [u'1.10.2']}

    Raises:
      :obj:`cnr.api.exception.ChannelNotFound`: channel not found

    See Also:
       * :obj:`cnr.api.registry.delete_channel_release`
    """
    channel = channel_class(name, package)
    channel.remove_release(release)
    return channel.to_dict()


def create_channel(package, name, channel_class):
    channel = channel_class(name, package)
    channel.save()
    return channel.to_dict()


def delete_channel(package, name, channel_class):
    channel = channel_class(name, package)
    channel.delete()
    return {"channel": channel.name, "package": package, "action": 'delete'}


def delete_package(package, version, package_class, channel_class):
    packagemodel = _get_package(package, version, package_class)
    package_class.delete(packagemodel.package, packagemodel.version, channel_class)
    return {"status": "delete", "package": packagemodel.package, "version": packagemodel.version}
